#!/usr/bin/env python3

import urllib3

urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)


from scilog import SciCat

url = "https://dacat.psi.ch/api/v3/"
cat = SciCat(url)

props = cat.proposals
nprops = len(props)
print(f"got {nprops} proposals")
print(props[0])



import functools
import logging

import requests
from requests import HTTPError

from .authmixin import HEADER_JSON, AuthError, AuthMixin

logger = logging.getLogger(__name__)

# TODO: potentially replace this module by the swagger API client solution, see
# https://github.com/paulscherrerinstitute/scilog/blob/testAutoGeneratedOpenAPI/sdk/python/python-client-generated/swagger_client/api_client.py


def authenticated(func):
    @functools.wraps(func)
    def authenticated_call(client, *args, **kwargs):
        if not isinstance(client, HttpClient):
            raise AttributeError("First argument must be an instance of HttpClient")
        if "headers" in kwargs:
            kwargs["headers"] = kwargs["headers"].copy()
        else:
            kwargs["headers"] = {}
        kwargs["headers"]["Authorization"] = client.token
        return func(client, *args, **kwargs)

    return authenticated_call


def formatted_http_error(func):
    @functools.wraps(func)
    def formatted_call(*args, **kwargs):
        try:
            out = func(*args, **kwargs)
        except HTTPError as exc:
            raise HTTPError(f"{exc.response.reason} Error Message: {exc.response.text}")
        return out

    return formatted_call


class HttpClient(AuthMixin):
    def __init__(self, *args, **kwargs):
        self._verify_certificate = True
        super().__init__(*args, **kwargs)
        self.login_path = self._login_path or (self.address + "/users/login")

    def authenticate(self, username, password):
        auth_payload = {"principal": username, "password": password}
        res = self._login(auth_payload, HEADER_JSON)
        try:
            token = "Bearer " + res["token"]
        except KeyError as e:
            raise AuthError(res) from e
        else:
            return token

    @authenticated
    @formatted_http_error
    def get_request(self, url, params=None, headers=None, timeout=10):
        logger.debug(f"Getting data request: {url} , {params}")
        response = requests.get(
            url, params=params, headers=headers, timeout=timeout, verify=self._verify_certificate
        )
        if response.ok:
            logger.debug(f"Getting data response: {response.json()}")
            return response.json()

        if response.reason == "Unauthorized":
            self.config.delete()
        raise response.raise_for_status()

    @authenticated
    @formatted_http_error
    def post_request(self, url, payload=None, files=None, headers=None, timeout=10):
        req = requests.post(
            url,
            json=payload,
            files=files,
            headers=headers,
            timeout=timeout,
            verify=self._verify_certificate,
        )
        req.raise_for_status()
        return req.json()

    @authenticated
    @formatted_http_error
    def patch_request(self, url, payload=None, files=None, headers=None, timeout=10):
        req = requests.patch(
            url,
            json=payload,
            files=files,
            headers=headers,
            timeout=timeout,
            verify=self._verify_certificate,
        )
        req.raise_for_status()
        return {}

    def _login(self, payload=None, headers=None, timeout=10):
        x = requests.post(
            self.login_path,
            json=payload,
            headers=headers,
            timeout=timeout,
            verify=self._verify_certificate,
        )
        return x.json()



import getpass
from abc import ABC, abstractmethod
from json.decoder import JSONDecodeError

from .config import Config

HEADER_JSON = {"Content-type": "application/json", "Accept": "application/json"}


def typename(obj):
    return type(obj).__name__


class AuthMixin(ABC):
    def __init__(self, address, options=None):
        self.address = address.rstrip("/")
        if not options:
            options = {}
        self._auto_save_token = options.get("auto_save_token", True)
        tn = typename(self).lower()
        try:
            self.config = Config(f".{tn}-tokens", auto_save_token=self._auto_save_token)
        except JSONDecodeError:
            self.config = {}

        self._token = options.get("token")
        self._username = options.get("username")
        self._password = options.get("password")
        self._login_path = options.get("login_path")
        self.options = options

    def __repr__(self):
        tn = typename(self)
        return f"{tn} @ {self.address}"

    @abstractmethod
    def authenticate(self, username, password):
        raise NotImplementedError

    @property
    def token(self):
        return self._retrieve_token()

    def reset_token(self):
        self._token = None

    def _retrieve_token(self):
        username = self._username or getpass.getuser()

        if self._token:
            # If a token is already set (either from options or from a previous authentication), use it.
            return self._token

        if not self._auto_save_token:
            # If auto-saving is disabled, authenticate without checking the config.
            self._token = self._authenticate()
            return self._token

        try:
            self._token = self.config[username]
        except KeyError:
            self._token = self._authenticate()
        self.config[username] = self._token
        return self._token

    def _authenticate(self):
        tn = typename(self)
        username = self._username or getpass.getuser()
        password = self._password or getpass.getpass(prompt=f"{tn} password for {username}: ")
        token = self.authenticate(username, password)
        return token


class AuthError(Exception):
    pass



from json import dumps
from unittest.mock import ANY, Mock, call, patch
from urllib.parse import quote

import pytest

from scilog import SciCat

ADDRESS = "http://scicat"

OPTIONS = {
    "username": f"username",
    "password": "password",
    "login_path": f"{ADDRESS}/login",
    "token_prefix": "Bearer ",
}


@pytest.fixture()
def scicat():
    scicat = SciCat(ADDRESS, options=OPTIONS)
    scicat.http_client.config = {}
    return scicat


@patch("requests.post")
@patch("requests.get")
def test_get_proposals(mock_get, mock_post, scicat):
    headers = {"Content-type": "application/json", "Accept": "application/json"}
    token = "token123"
    mock_response = Mock()
    mock_response.json.return_value = {"id": token}
    mock_post.return_value = mock_response
    list(scicat.proposals)
    mock_post.assert_called_with(
        OPTIONS["login_path"],
        json={"username": OPTIONS["username"], "password": OPTIONS["password"]},
        headers=headers,
        timeout=ANY,
        verify=True,
    )
    filter = {"limits": {"skip": 0, "limit": 500}}
    mock_get.assert_called_with(
        f"{ADDRESS}/proposals?filters={quote(dumps(filter))}",
        params=None,
        headers={**headers, "Authorization": f"{OPTIONS['token_prefix']}{token}"},
        timeout=ANY,
        verify=True,
    )


@patch("scilog.scicat.SciCatRestAPI.get_request")
def test__proposals_batch(mock_get, scicat):

    mock_get.side_effect = [[1, 2], [3, 4], []]
    filters = [{"limits": {"skip": 0, "limit": 500}}, {"limits": {"skip": 500, "limit": 500}}]
    for _ in scicat._proposals_batch():
        continue

    assert mock_get.call_count == 3

    expected_calls = [
        call(f"{ADDRESS}/proposals?filters={quote(dumps(filters[0]))}", headers=ANY),
        call(f"{ADDRESS}/proposals?filters={quote(dumps(filters[1]))}", headers=ANY),
    ]
    mock_get.assert_has_calls(expected_calls, any_order=False)


@patch("scilog.scicat.SciCatRestAPI.get_request")
@pytest.mark.parametrize(
    "return_options",
    [
        None,
        {"lazy": True},
        {"lazy": False},
    ],
)
def test_proposals(mock_get, return_options):
    scicat = SciCat(ADDRESS, options=OPTIONS, return_options=return_options)
    scicat.http_client.config = {}
    mock_get.side_effect = [[1, 2], [3, 4], []]
    proposals = [1, 2, 3, 4]
    scicat_proposals = scicat.proposals
    for i, p in enumerate(scicat_proposals):
        assert p == proposals[i]
    lazy = return_options.get("lazy", False) if return_options else False
    assert len(list(scicat_proposals)) == (0 if lazy else len(proposals))



from unittest.mock import ANY, Mock, patch

import pytest

from scilog import SciCat


@patch("requests.post")
@patch("requests.get")
@pytest.mark.parametrize(
    "token_prefix",
    [
        "",
        None,
    ],
)
def test_get_proposals(mock_get, mock_post, token_prefix):
    address = "http://scicat"
    options = {
        "username": f"username{token_prefix}",
        "password": "password",
        "login_path": f"{address}/login",
        "token_prefix": token_prefix,
    }
    headers = {"Content-type": "application/json", "Accept": "application/json"}
    token = "token123"

    scicat = SciCat(address, options=options)
    mock_response = Mock()
    mock_response.json.return_value = {"id": token}
    mock_post.return_value = mock_response
    scicat.http_client.config = {}
    scicat.proposals
    mock_post.assert_called_with(
        options["login_path"],
        json={"username": options["username"], "password": options["password"]},
        headers=headers,
        timeout=ANY,
        verify=True,
    )
    mock_get.assert_called_with(
        f"{address}/proposals",
        params=None,
        headers={**headers, "Authorization": f"{token_prefix or ''}{token}"},
        timeout=ANY,
        verify=True,
    )



import json
from pathlib import Path


class Config(dict):
    def __init__(self, fname, folder=None, auto_save_token=True):
        if folder is not None:
            folder = Path(folder)
        else:
            folder = Path.home()
        self.fname = folder / fname
        self.auto_save_token = auto_save_token
        content = self._load()
        super().__init__(content)

    def __setitem__(self, name, value):
        self.update(**{name: value})

    def update(self, **kwargs):
        super().update(**kwargs)
        self._save()

    def _load(self):
        fn = self.fname
        if fn.exists():
            return json_load(fn)
        else:
            return {}

    def _save(self):
        json_save(self, self.fname)

    def delete(self):
        if self.fname.exists():
            self.fname.unlink()


def json_save(what, filename, *args, indent=4, sort_keys=True, **kwargs):
    with open(filename, "w") as f:
        json.dump(what, f, *args, indent=indent, sort_keys=sort_keys, **kwargs)


def json_load(filename, *args, **kwargs):
    with open(filename, "r") as f:
        return json.load(f, *args, **kwargs)



from unittest import mock

import pytest

from scilog import LogbookMessage


@pytest.mark.parametrize(
    "input_text,out_textcontent",
    [([""], "<p></p>"), (["text 1", "text 2"], "<p>text 1</p><p>text 2</p>")],
)
def test_add_text(input_text, out_textcontent):
    msg = LogbookMessage()
    for text in input_text:
        msg.add_text(text)
    assert msg._content.textcontent == out_textcontent


def test_add_file():
    msg = LogbookMessage()
    with pytest.raises(FileNotFoundError):
        msg.add_file("test_file.png")

    with mock.patch("os.path.exists", return_value=True):
        msg.add_file("test_file.png")
        assert msg._content.files[0]["filepath"] == "test_file.png"
        msg.add_file("test_file2.png")
        assert msg._content.files[1]["filepath"] == "test_file2.png"


@pytest.mark.parametrize("tags", ["tag1", ["tag1", "tag2"]])
def test_add_tag(tags):
    msg = LogbookMessage()
    msg.add_tag(tags)
    if not isinstance(tags, list):
        tags = [tags]
    assert set(msg._content.tags) & set(tags)
    msg.add_tag("tag3")
    assert msg._content.tags[-1] == "tag3"



from __future__ import annotations

import functools
import json
import logging
import os
import uuid
import warnings
from typing import Tuple, TypeVar, cast

from typeguard import typechecked

import scilog.logbook_message as lm
from .authmixin import HEADER_JSON
from .httpclient import HttpClient
from .models import Basesnippet, Filesnippet, Location, Logbook, Paragraph, snippet_factory

T = TypeVar("T", bound="Basesnippet")

ACLS = ["createACL", "readACL", "updateACL", "deleteACL", "shareACL", "adminACL"]


logger = logging.getLogger(__name__)


def pinned_to_logbook(logbook_keys, include_none=False):
    """
    Decorator to pin methods to the currently selected logbook.
    This means that the decorated method will automatically include
    the logbook's id in the specified keys (e.g., parentId) of the payload,
    if they are not already provided.
    """

    def pinned_to_logbook_inner(func):
        @functools.wraps(func)
        def pinned_to_logbook_call(log, *args, **kwargs):
            if not isinstance(log.logbook, Basesnippet):
                warnings.warn("No logbook selected.")
                return func(log, *args, **kwargs)

            logbook = log.logbook.model_dump(exclude_none=not include_none)
            logger.debug(f"====== pinned to logbook: {kwargs}")
            for key in logbook_keys:
                logger.debug(f"======pinned to logbook: {key}")
                if key in kwargs:
                    continue
                if key == "parentId":  # mapping from childinstance to parentinstance
                    if "where" not in kwargs:
                        kwargs[key] = log.logbook.id
                    else:
                        kwargs["where"] = {"and": [kwargs["where"], {"parentId": log.logbook.id}]}
                else:
                    logger.debug(f"Non parentId condition in pinned_to_logbook: {key}")
                    if logbook.get(key):
                        kwargs[key] = logbook[key]
            logger.debug(f"========= pinned to logbook: resulting kwargs: {kwargs}")
            return func(log, *args, **kwargs)  # real call happens here

        return pinned_to_logbook_call

    return pinned_to_logbook_inner


class SciLogRestAPI(HttpClient):
    pass


class SciLogCore:
    """
    Core functionalities of SciLog, including methods for posting snippets and locations,
    uploading files, and patching snippets. This class is intended to be used internally by
    the SciLog class and is not meant to be accessed directly by users.
    """

    IMAGE_TYPES = ["png", "jpg", "jpeg"]

    def __init__(self, high_level_interface) -> None:
        self._hli = high_level_interface

    @property
    def logbook(self):
        return self._hli.logbook

    @logbook.setter
    def logbook(self, val):
        self._hli.logbook = val

    @property
    def http_client(self):
        return self._hli.http_client

    def make_filter(
        self,
        where: dict | None = None,
        limit: int = 0,
        skip: int = 0,
        fields: dict | None = None,
        include: dict | None = None,
        order: list | None = None,
    ) -> dict:
        """
        Create a filter dictionary for querying snippets.

        Args:
            where (dict | None): Conditions to filter the snippets.
            limit (int): Maximum number of snippets to return.
            skip (int): Number of snippets to skip.
            fields (dict | None): Specific fields to include or exclude.
            include (dict | None): Related models to include.
            order (list | None): Order in which to return the snippets.

        Returns:
            dict: A dictionary representing the filter.

        """
        filt = {}
        if where is not None:
            # TOOD needed ? items = [where.copy()]
            filt["where"] = where  #
        if limit > 0:
            filt["limit"] = limit
        if skip > 0:
            filt["skip"] = skip
        if fields is not None:
            filt["fields"] = fields
        if include is not None:
            filt["include"] = include
        if order is not None:
            filt["order"] = order
        filt = json.dumps(filt)
        return {"filter": filt}

    # TODO if this is a user function add docstring
    @pinned_to_logbook(["parentId", *ACLS])
    def post_snippet(self, **kwargs) -> Paragraph:
        """
        Post a new snippet.

        Args:
            **kwargs: Keyword arguments to be included in the snippet payload.

        Returns:
            Paragraph: The newly created snippet.
        """
        url = self.http_client.address + "/basesnippets"
        payload = kwargs
        if payload.get("files"):
            payload = self.upload_files(payload)

        out = cast(
            Paragraph,
            Paragraph.from_http_response(
                self.http_client.post_request(url, payload=payload, headers=HEADER_JSON)
            ),
        )
        return out

    # TODO if this is a user function add docstring
    @pinned_to_logbook(["parentId", *ACLS])
    def post_location(self, **kwargs) -> Location:
        """
        Post a new location snippet.

        Args:
            **kwargs: Keyword arguments to be included in the Location payload.

        Returns:
            Location: The newly created Location snippet.
        """
        url = self.http_client.address + "/basesnippets"
        payload = kwargs
        if payload.get("files"):
            payload = self.upload_files(payload)
        out = cast(
            Location,
            Location.from_http_response(
                self.http_client.post_request(url, payload=payload, headers=HEADER_JSON)
            ),
        )
        return out

    def upload_files(self, payload: dict) -> dict:
        """
        Upload files specified in the payload and update the payload with
        the returned fileIds and accessHashes. Typically not used directly
        by the user, but rather as part of the post_snippet and post_location methods.

        Args:
            payload (dict): The original payload containing file paths.

        Returns:
            dict: The updated payload with fileIds and accessHashes.
        """
        for file in payload.get("files", []):
            if "fileId" in file and file["fileId"] is not None:
                continue
            logger.debug(f"Posting from filepath: {file['filepath']}")
            filesnippet = self._post_filesnippet(
                file["filepath"],
                **{key: val for key, val in payload.items() if key in [*ACLS, "tags"]},
            )
            file["fileId"] = filesnippet.id
            file["accessHash"] = filesnippet.accessHash
            file.pop("filepath")
        return payload

    def _post_filesnippet(self, filepath: str, **kwargs) -> Filesnippet:
        """
        Upload a file as a Filesnippet. Primarily used by the
        upload_files method to handle file uploads for snippets that contain files.

        Args:
            filepath (str): Path to the file that ought to be uploaded
            **kwargs: Additional keyword arguments to be included in the Filesnippet payload

        Raises:
            FileNotFoundError: Raised if the file specified in filepath does not exist
            ValueError: Raised if the filepath is not pointing to a file

        Returns:
            Filesnippet: Filesnippet containing the metadata of the newly created entry
        """
        url = self.http_client.address + "/filesnippet/files"

        file_extension = os.path.splitext(filepath)[-1]
        if not file_extension:
            raise ValueError("filepath must be pointing to a file, not a directory.")
        file_extension = file_extension[1:]

        snippet = Filesnippet(**kwargs)
        snippet.fileExtension = file_extension
        payload = snippet.model_dump(exclude_none=True)

        file_descriptor = "image" if file_extension in self.IMAGE_TYPES else "file"
        multipart_form_data = {
            "file": (
                filepath + "." + file_extension,
                open(filepath, "rb"),
                f"{file_descriptor}/{file_extension}",
            ),
            "fields": (None, json.dumps(payload)),
        }
        obj = Filesnippet.from_http_response(
            self.http_client.post_request(
                url, files=multipart_form_data, headers={"Accept": "application/json"}
            )
        )
        if isinstance(obj, list):
            return obj[0]
        return obj

    @pinned_to_logbook(ACLS)
    def post_file(self, filepath: str, **kwargs) -> Filesnippet:
        """
        Upload a file. As they are disconnected from paragraphs and logbooks, there
        is no need to copy the parentId. Instead we simply keep the ACLS.

        Args:
            filepath (str): Path to the file that ought to be uploaded

        Raises:
            FileNotFoundError: Raised if the file specified in filepath does not exist
            ValueError: Raised if the filepath is not pointing to a file

        Returns:
            Filesnippet: Filesnippet containing the metadata of the newly created entry
        """
        if not os.path.exists(filepath):
            raise FileNotFoundError("Specified file does not exist.")

        fsnippet = self._post_filesnippet(filepath, **kwargs)
        # ret = self._file_upload(filepath, fsnippet.id, file_extension)
        return fsnippet

    @pinned_to_logbook(ACLS)
    def append_files_to_snippet(
        self, snippet: Paragraph, filepaths: list[str], **kwargs
    ) -> Paragraph:
        """
        Append files or images to an already existing snippet. Files and images will be
        appended following the order given in 'filepaths'. As the files / images are disconnected
        from paragraphs and logbooks, there is no need to copy the parentId. Instead we simply
        keep the ACLS.

        Args:
            snippet (Paragraph): Snippet to which the files should be appended
            filepaths (list): List of file paths pointing to the files that should be uploaded

        Returns:
            Paragraph: Updated snippet
        """
        for filepath in filepaths:
            fsnippet = self.post_file(filepath)

            if not isinstance(snippet.files, list):
                snippet.files = []

            # if we reach this point, we can assume that filepath has been checked (cf. self.post_file)
            file_info, file_textcontent = self.prepare_file_content(
                filepath=filepath, fsnippet=fsnippet
            )
            if not snippet.textcontent:
                snippet.textcontent = ""
            snippet.textcontent += file_textcontent
            snippet.files.append(file_info)

        return self.patch_snippet(snippet)

    @staticmethod
    def prepare_file_content(
        filepath: str, fsnippet: Filesnippet | None = None
    ) -> Tuple[dict, str]:
        """
        Prepare file content and textcontent for appending to a snippet.

        Args:
            filepath (str): Path to the file that ought to be prepared
            fsnippet (Filesnippet | None): Optional Filesnippet containing metadata of the file

        Returns:
            Tuple[dict, str]: A tuple containing the file information dictionary and the text content string
        """
        file_extension = filepath.split(".")[-1].lower()
        file_hash = str(uuid.uuid4())
        file_id = fsnippet.id if fsnippet and fsnippet.id else None
        access_hash = fsnippet.accessHash if fsnippet and fsnippet.accessHash else None

        if file_extension in SciLogCore.IMAGE_TYPES:
            textcontent = (
                f'<figure class="image image_resized"><img src="" title="{file_hash}"></figure>'
            )
            file_info = {
                "fileHash": file_hash,
                "filepath": filepath,
                "fileExtension": f"image/{file_extension}",
                "fileId": file_id,
                "accessHash": access_hash,
                "style": {"width": "82.25%", "height": ""},
            }

        else:
            textcontent = f'<p><a class="fileLink" target="_blank" href="file:{file_hash}">{os.path.basename(filepath)}</a></p>'
            file_info = {
                "fileHash": file_hash,
                "filepath": filepath,
                "fileExtension": f"file/{file_extension}",
                "fileId": file_id,
                "accessHash": access_hash,
            }
        return (file_info, textcontent)

    def patch_snippet(self, snippet: T, **kwargs) -> T:
        """Update (patch) snippet with given snippet.
        Args:
            snippet (T): Snippet containing the newly updated fields
        Returns:
            T: Updated snippet
        """
        url = self.http_client.address + "/basesnippets/" + snippet.id
        snippet.id = None
        snippet.createdAt = None
        snippet.createdBy = None
        snippet.updatedAt = None
        snippet.updatedBy = None
        snippet.expiresAt = None
        payload = snippet.model_dump(exclude_none=True)

        out = cast(
            T,
            snippet_factory(
                self.http_client.patch_request(url, payload=payload, headers=HEADER_JSON)
            ),
        )
        return out


class SciLog:
    def __init__(self, address: str, options: dict | None = None, **kwargs):
        """
        Initialize the SciLog client.

        Args:
            address (str): The base URL of the SciLog server, e.g. "https://scilog.psi.ch/api/v1"
            options (dict | None): Optional dictionary containing authentication options such as
                    "token", "username", "password", "login_path" and "auto_save_token". If not
                    provided, the client will attempt to retrieve these from the environment or
                    prompt the user.
            **kwargs: Additional keyword arguments for authentication, if needed.
        """
        self.http_client = SciLogRestAPI(address=address, options=options, **kwargs)
        self.logbook = None
        self.core = SciLogCore(self)

    def select_logbook(self, logbook: Logbook):
        """
        Select a logbook to work with. This will pin subsequent operations to the selected logbook.

        Args:
            logbook (Logbook): The logbook to select.
        """
        self.logbook = logbook

    def new(self, text: str | None = None) -> lm.LogbookMessage:
        """
        Create a new LogbookMessage instance, optionally with initial text content.

        Args:
            text (str | None): Optional initial text content for the LogbookMessage.

        Returns:
            LogbookMessage: A new instance of LogbookMessage with the provided text content.

        Raises:
            ValueError: If no logbook is currently selected.
        """
        if self.logbook is None:
            raise ValueError(
                "No logbook selected. Please select a logbook before creating a message."
            )
        msg = lm.LogbookMessage()
        if text:
            msg.add_text(text)
        return msg

    @pinned_to_logbook(["parentId"])
    def get_snippets(
        self,
        where: dict | None = None,
        limit: int = 0,
        skip: int = 0,
        fields: dict | None = None,
        include: dict | None = None,
        order: list | None = None,
        **kwargs,
    ) -> list[Basesnippet]:
        """
        Retrieve snippets from the selected logbook.

        Args:
            where (dict | None): Filter conditions for the query.
            limit (int): Maximum number of snippets to retrieve.
            skip (int): Number of snippets to skip.
            fields (dict | None): Fields to include in the response.
            include (dict | None): Related entities to include in the response.
            order (list | None): Order in which to return the snippets.
            **kwargs: Additional keyword arguments for filtering, if 'where' is not provided.

        Returns:
            list[Basesnippet]: A list of retrieved snippets.
        """
        url = self.http_client.address + "/basesnippets"

        if not where:
            where = {"and": [kwargs.copy()]}
            logger.debug("==== where condition defined via keyword arguments:%s", where)
        else:
            logger.debug("==== where condition defined directly:%s", where)
        params = self.core.make_filter(
            where=where, limit=limit, skip=skip, fields=fields, include=include, order=order
        )
        # TODO dont allow case that both where and kwargs are defined, return error in this case
        out = cast(
            list[Basesnippet],
            snippet_factory(self.http_client.get_request(url, params=params, headers=HEADER_JSON)),
        )
        return out

    @pinned_to_logbook(["parentId", *ACLS])
    def import_from_dict(self, snippet: dict, **kwargs):
        msg = snippet["textcontent"]
        snippet_dict = {k: v for k, v in snippet.items() if k != "textcontent"}
        self.send_message(msg, **snippet_dict)

    @pinned_to_logbook(["parentId", *ACLS])
    def send_message(self, msg: str, **kwargs):
        """
        Send a message to the selected logbook.

        For more complex message compositions, it is recommended to use the send_logbook_message method,
        which allows for more flexibility and control over the message content and formatting.

        Args:
            msg (str): The message text to send.
            **kwargs: Additional keyword arguments to be included in the message payload.

        """
        lm_msg = lm.LogbookMessage(**kwargs)
        lm_msg.add_text(msg)

        self.send_logbook_message(lm_msg)

    @typechecked
    def send_logbook_message(self, msg: lm.LogbookMessage) -> None:
        """Upload a new message to SciLog.

        Args:
            msg (lm.LogbookMessage): Previously composed LogbookMessage.

        Example:
            >>> msg = LogbookMessage()
            >>> msg.add_text("Another example text").add_tag(["color"])
            >>> log.send_logbook_message(msg)

        """
        payload = msg._content.model_dump(exclude_none=True)
        payload["linkType"] = "paragraph"
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self.core.post_snippet(**payload)

    def get_logbooks(
        self,
        where: dict | None = None,
        limit: int = 0,
        skip: int = 0,
        fields: dict | None = None,
        include: dict | None = None,
        order: list | None = None,
        **kwargs,
    ) -> list[Logbook]:
        """
        Retrieve logbooks available to the user, with optional filtering and pagination.

        Args:
            where (dict | None): Filter conditions for the query.
            limit (int): Maximum number of logbooks to retrieve.
            skip (int): Number of logbooks to skip.
            fields (dict | None): Fields to include in the response.
            include (dict | None): Related entities to include in the response.
            order (list | None): Order in which to return the logbooks.
            **kwargs: Additional keyword arguments for filtering, if 'where' is not provided.

        Returns:
            list[Logbook]: A list of retrieved logbooks.
        """

        url = self.http_client.address + "/logbooks"

        if not where:
            where = {"and": [kwargs.copy()]}
            logger.debug("==== where condition defined via keyword arguments:%s", where)
        else:
            logger.debug("==== where condition defined directly:%s", where)

        params = self.core.make_filter(
            where=where, limit=limit, skip=skip, fields=fields, include=include, order=order
        )
        out = cast(
            list[Logbook],
            snippet_factory(self.http_client.get_request(url, params=params, headers=HEADER_JSON)),
        )
        return out

    # TODO: review and potentially move the whole core logic here
    def post_snippet(self, **kwargs) -> Paragraph:
        """
        Post a new snippet. This method is pinned to the currently selected logbook, meaning that
        the logbook's id will automatically be included in the payload.

        Note: For sending messages to the logbook, it is recommended to use the
        send_logbook_message method, which is specifically designed for that purpose and
        provides a more user-friendly interface.

        Args:
            **kwargs: Keyword arguments to be included in the snippet payload.

        Returns:
            Paragraph: The newly created snippet.
        """
        return self.core.post_snippet(**kwargs)

    # TODO: review and potentially move the whole core logic here
    def post_location(self, **kwargs) -> Location:
        """
        Post a new location. This method is pinned to the currently selected logbook, meaning that
        the logbook's id will automatically be included in the payload.

        Args:
            **kwargs: Keyword arguments to be included in the location payload.

        Returns:
            Location: The newly created location.
        """
        return self.core.post_location(**kwargs)
